#!/bin/sh
# the next line restarts using wish (must be wish4.0 or newer) \
exec wish "$0" -- "$@"

# $Id: puzzle.tcl,v 1.3 1996/11/17 20:41:02 uwe Exp $

source setopt.tcl
source evaloptions.tcl
source vector.tcl
source forallfields.tcl
source drawstatus.tcl
source preparepictures.tcl
source readpictures.tcl

# only for restart
catch {unset puzzle}
catch {unset images}

# define index entry
# and set drawField as view routine
proc initField {ix args} {
    global puzzle

    set puzzle($ix) 0
    trace variable puzzle($ix) w drawField

}

proc initAllFields {} {
    forallFields initField
}

proc createPuzzleCanvas {} {
    global puzzle

    set w $puzzle(window)

    catch {destroy $w}

    canvas $w \
	-width	$puzzle(canvasSize) \
	-height	$puzzle(canvasSize) \
	-relief flat \
	-bd 0 \
	-highlightthickness 0

    pack $w

}

proc setField {ix pic} {
    global puzzle 

    set puzzle([list [expr $ix % 3] [expr $ix / 3]]) $pic
    update
}

proc clearAllFields {} {
    for {set i 0} {$i < 9} {incr i} {setField $i -1}
}

proc solveAutomatic {} {
    global puzzle

    set file "$puzzle(pictureFile)-solution.tcl"
    if {[file isfile $file]} {
	clearAllFields
	source $file
    }
}

proc drawField {var index w} {
    global puzzle images

    set w    $puzzle(window)
    set pos  [eval scaleVec $index $puzzle(fieldSize)]
    set pos1 [eval scaleVec [eval addVec $index {1 1}] $puzzle(fieldSize)]
    set tag "pic-[lindex $index 0]-[lindex $index 1]"
    set img $images($puzzle($index))

    catch {$w delete $tag}
    eval [list $w create image] $pos [list -image $img -anchor nw -tag $tag]
    eval [list $w create rectangle] $pos $pos1 [list -fill {} -outline black -tag $tag]

    # message "drawField: puzzle($index) = $puzzle($index)"

    drawStatus
}

proc computeIndex {x y} {
    global puzzle

    set fs  $puzzle(fieldSize)
    set row [expr $x / $fs]
    set col [expr $y / $fs]

    return [list $row $col]
}

proc computeField {x y} {
    global puzzle

    set fs  $puzzle(fieldSize)
    set row [expr $x / $fs]
    set col [expr $y / $fs]

    set start [list $x $y]
    set org   [list [expr $row * $fs + $fs / 2] [expr $col * $fs + $fs / 2]]

    # default direction of vector is up
    if {$start == $org} {
	set start [eval addVec $start 0 -1]
    }

    set puzzle(currField) [list $row $col]
    set puzzle(currStart) $start
    set puzzle(currOrg)   $org

    set puzzle(currRot)   [eval rotMat [eval subVec $start $org]]
    set puzzle(currFlips) 0

    drawStatus
}

# compute angle between currStart and vector {x y}
proc computeAngle {x y} {
    global puzzle

    if [info exists puzzle(currStart)] {
	set p [list $x $y]
	set org $puzzle(currOrg)
	if {$p == $org} {
	    return 0
	}
	set p0 [eval subVec $p $org]
	set p1 [eval multVec $puzzle(currRot) $p0]
	set angle [eval 2angleVec $p1]

	# message "computeAngle: $angle"
	return $angle
    }
    return 0
}

# compute the rotation clock counterwise in 90 degree units
proc computeRotation {x y} {
    set angle [computeAngle $x $y]
    set rot [expr ( round($angle) + 45 ) / 90 % 4]

    # message "computeRotation: rotate $rot steps"
    return $rot
}

# compute, whether picture must be rotated
proc computeFlips {x y} {
    global puzzle

    set newflips [computeRotation $x $y]
    set curflips $puzzle(currFlips) 
    if {$newflips != $curflips} {
	set puzzle(currFlips) $newflips
	flipField $puzzle(currField) [expr ($newflips + 4 - $curflips) % 4]
    }
}

# rotate field
proc flipField {field steps} {
    global puzzle

    set nr  [expr $puzzle($field) / 4 * 4]
    set rot [expr $puzzle($field) - $nr]

    set puzzle($field) [expr $nr + ( $rot + $steps ) % 4]
}

proc computeSwaps {x y} {
    global puzzle

    set newix [computeIndex $x $y]
    set oldix $puzzle(currField)
    if {$newix != $oldix} {
	set t $puzzle($oldix)
	set puzzle($oldix) $puzzle($newix)
	set puzzle($newix) $t
	computeField $x $y
    }
}

# ------------------------------------------------------------------------------
#

proc setPuzzleBindings {} {
    global puzzle

    set w $puzzle(window)

    bind $w <Button-2>  {computeField %x %y}
    bind $w <B2-Motion> {computeFlips %x %y}
    bind $w <Button-1>  {computeField %x %y}
    bind $w <B1-Motion> {computeSwaps %x %y}
    bind $w q           {destroy .}
    bind $w s           {solveAutomatic}

    focus $w
}

# ------------------------------------------------------------------------------
#

eval options $argv

initAllFields

createPuzzleCanvas

setPuzzleBindings

preparePictures

readAllPictures

set puzzle(drawStatus) $puzzle(debug)